import gzip
import struct
import pathlib
import warnings
import functools
from pathlib import Path
from collections import OrderedDict
from urllib.request import Request

import cdflib
import numpy as np
import pandas as pd
from scipy.io import readsav

import astropy.units as u
from astropy.io import fits
from astropy.io.fits import Header
from astropy.time import Time
from sunpy import log
from sunpy.data import cache
from sunpy.net import attrs as a
from sunpy.time import parse_time
from sunpy.util.datatype_factory_base import (
    BasicRegistrationFactory,
    MultipleMatchError,
    NoMatchError,
    ValidationFunctionError,
)
from sunpy.util.exceptions import SunpyUserWarning, warn_user
from sunpy.util.io import is_url, parse_path, possibly_a_path
from sunpy.util.metadata import MetaDict
from sunpy.util.util import expand_list

from radiospectra.exceptions import NoSpectrogramInFileError, SpectraMetaValidationError
from radiospectra.spectrogram.spectrogrambase import GenericSpectrogram
from radiospectra.utils import subband_to_freq

SUPPORTED_ARRAY_TYPES = (np.ndarray,)
try:
    import dask.array

    SUPPORTED_ARRAY_TYPES += (dask.array.Array,)
except ImportError:
    pass

__all__ = ["SpectrogramFactory", "Spectrogram"]


class SpectrogramFactory(BasicRegistrationFactory):
    """
    A factory for generating spectrograms.

    Parameters
    ----------
    \\*inputs
        `str` or `pathlib.Path` to the file.

    Returns
    -------
    `radiospectra.spectrogram.Spectrogram`
        The spectrogram for the give file
    """

    def _validate_meta(self, meta):
        """
        Validate a meta argument.
        """
        if isinstance(meta, Header):
            return True
        elif isinstance(meta, dict):
            return True
        else:
            return False

    def _parse_args(self, *args, silence_errors=False, **kwargs):
        """
        Parses an args list into data-header pairs.

        args can contain any mixture of the following
        entries:

        * tuples of data,header
        * data, header not in a tuple
        * data, wcs object in a tuple
        * data, wcs object not in a tuple
        * filename, as a str or pathlib.Path, which will be read
        * directory, as a str or pathlib.Path, from which all files will be read
        * glob, from which all files will be read
        * url, which will be downloaded and read
        * lists containing any of the above.

        Example
        -------
        self._parse_args(data, header,
                         (data, header),
                         ['file1', 'file2', 'file3'],
                         'file4',
                         'directory1',
                         '*.fits')
        """
        # Account for nested lists of items
        args = expand_list(args)
        # Sanitize the input so that each 'type' of input corresponds to a different
        # class, so single dispatch can be used later
        nargs = len(args)
        i = 0
        while i < nargs:
            arg = args[i]
            if isinstance(arg, SUPPORTED_ARRAY_TYPES):
                # The next two items are data and a header
                data = args.pop(i)
                header = args.pop(i)
                args.insert(i, (data, header))
                nargs -= 1
            elif isinstance(arg, str) and is_url(arg):
                # Repalce URL string with a Request object to dispatch on later
                args[i] = Request(arg)
            elif possibly_a_path(arg):
                # Repalce path strings with Path objects
                args[i] = pathlib.Path(arg)
            i += 1
        # Parse the arguments
        # Note that this list can also contain GenericMaps if they are directly given to the factory
        data_header_pairs = []
        for arg in args:
            try:
                data_header_pairs += self._parse_arg(arg, **kwargs)
            except NoSpectrogramInFileError as e:
                if not silence_errors:
                    raise
                warn_user(f"One of the arguments failed to parse with error: {e}")
        return data_header_pairs

    @functools.singledispatchmethod
    def _parse_arg(self, arg, **kwargs):
        """
        Take a factory input and parse into (data, header) pairs.

        Must return a list, even if only one pair is returned.
        """
        raise ValueError(f"Invalid input: {arg}")

    @_parse_arg.register(tuple)
    def _parse_tuple(self, arg, **kwargs):
        # Data-header
        data, header = arg
        pair = data, header
        if self._validate_meta(header):
            pair = (data, OrderedDict(header))
        return [pair]

    @_parse_arg.register(GenericSpectrogram)
    def _parse_map(self, arg, **kwargs):
        return [arg]

    @_parse_arg.register(Request)
    def _parse_url(self, arg, **kwargs):
        url = arg.full_url
        path = str(cache.download(url).absolute())
        pairs = self._read_file(path, **kwargs)
        return pairs

    @_parse_arg.register(pathlib.Path)
    def _parse_path(self, arg, **kwargs):
        return parse_path(arg, self._read_file, **kwargs)

    def __call__(self, *args, silence_errors=False, **kwargs):
        """
        Method for running the factory.

        Takes arbitrary arguments and keyword arguments and passes
        them to a sequence of pre-registered types to determine which is the correct spectrogram-
        type to build. Arguments args and kwargs are passed through to the validation function and
        to the constructor for the final type. For spectrogram types, validation function must take
        a data-header pair as an argument.

        Parameters
        ----------
        silence_errors : `bool`, optional
            If set, ignore data-header pairs which cause an exception.
            Default is ``False``.

        Notes
        -----
        Extra keyword arguments are passed through to `sunpy.io.read_file` such
        as `memmap` for FITS files.
        """
        data_header_pairs = self._parse_args(*args, silence_errors=silence_errors, **kwargs)
        new_maps = list()
        # Loop over each registered type and check to see if WidgetType
        # matches the arguments.  If it does, use that type.
        for pair in data_header_pairs:
            if isinstance(pair, GenericSpectrogram):
                new_maps.append(pair)
                continue
            data, header = pair
            meta = MetaDict(header)
            try:
                new_map = self._check_registered_widgets(data, meta, **kwargs)
                new_maps.append(new_map)
            except (NoMatchError, MultipleMatchError, ValidationFunctionError, SpectraMetaValidationError) as e:
                if not silence_errors:
                    raise
                warnings.warn(f"One of the data, header pairs failed to validate with: {e}", SunpyUserWarning)
        if not len(new_maps):
            raise RuntimeError("No maps loaded")
        if len(new_maps) == 1:
            return new_maps[0]
        return new_maps

    def _check_registered_widgets(self, data, meta, **kwargs):
        candidate_widget_types = list()
        for key in self.registry:
            # Call the registered validation function for each registered class
            if self.registry[key](data, meta, **kwargs):
                candidate_widget_types.append(key)

        n_matches = len(candidate_widget_types)
        if n_matches == 0:
            if self.default_widget_type is None:
                raise NoMatchError("No types match specified arguments and no default is set.")
            else:
                candidate_widget_types = [self.default_widget_type]
        elif n_matches > 1:
            raise MultipleMatchError(
                "Too many candidate types identified "
                f"({candidate_widget_types}). "
                "Specify enough keywords to guarantee unique type "
                "identification."
            )

        # Only one is found
        WidgetType = candidate_widget_types[0]
        return WidgetType(data, meta, **kwargs)

    def _read_file(self, file, **kwargs):
        file = Path(file)
        extensions = file.suffixes
        first_extension = extensions[0].lower()
        if first_extension == ".dat":
            return self._read_dat(file)
        elif first_extension in (".r1", ".r2"):
            return [self._read_idl_sav(file, instrument="waves")]
        elif first_extension == ".cdf":
            cdf = self._read_cdf(file)
            if isinstance(cdf, list):
                return cdf
            return [cdf]
        elif first_extension == ".srs":
            return [self._read_srs(file)]
        elif first_extension in (".fits", ".fit", ".fts", "fit.gz"):
            return [self._read_fits(file)]
        else:
            raise ValueError(f"Extension {extensions[0]} not supported.")

    @staticmethod
    def _read_dat(file):
        if "swaves" in file.name:
            name, prod, date, spacecraft, receiver = file.stem.split("_")
            # frequency range
            freqs = np.genfromtxt(file, max_rows=1) * u.kHz
            # bg which is already subtracted from data
            bg = np.genfromtxt(file, skip_header=1, max_rows=1)
            # data
            data = np.genfromtxt(file, skip_header=2)
            times = data[:, 0] * u.min
            data = data[:, 1:].T
            meta = {
                "instrument": name,
                "observatory": f"STEREO {spacecraft.upper()}",
                "product": prod,
                "start_time": Time.strptime(date, "%Y%m%d"),
                "wavelength": a.Wavelength(freqs[0], freqs[-1]),
                "detector": receiver,
                "freqs": freqs,
                "background": bg,
            }
            meta["times"] = meta["start_time"] + times
            meta["end_time"] = meta["start_time"] + times[-1]
            return data, meta
        elif "bst" in file.name:
            subbands = (np.arange(54, 454, 2), np.arange(54, 454, 2), np.arange(54, 230, 2))
            num_subbands = 488

            data = np.fromfile(file)
            polarisation = file.stem[-1]

            num_times = data.shape[0] / num_subbands
            if not num_times.is_integer():
                log.warning("BST file seems incomplete dropping incomplete frequencies")
                num_times = np.floor(num_times).astype(int)
                truncate = num_times * num_subbands
                data = data[:truncate]
            data = data.reshape(-1, num_subbands).T  # (Freq x Time).T = (Time x Freq)
            dt = np.arange(num_times) * 1 * u.s
            start_time = Time.strptime(file.name.split("_bst")[0], "%Y%m%d_%H%M%S")
            times = start_time + dt

            obs_mode = (3, 5, 7)

            freqs = [subband_to_freq(sb, mode) for sb, mode in zip(subbands, obs_mode)]

            # 1st 200 sbs mode 3, next 200 sbs mode 5, last 88 sbs mode 7
            spec = {0: data[:200, :], 1: data[200:400, :], 2: data[400:, :]}
            data_header_pairs = []
            for i in range(3):
                meta = {
                    "instrument": "ILOFAR",
                    "observatory": "Birr (IE613)",
                    "start_time": times[0],
                    "mode": obs_mode[i],
                    "wavelength": a.Wavelength(freqs[i][0], freqs[i][-1]),
                    "freqs": freqs[i],
                    "times": times,
                    "end_time": times[-1],
                    "detector": "ILOFAR",
                    "polarisation": polarisation,
                }

                data_header_pairs.append((spec[i], meta))
            return data_header_pairs
        else:
            raise ValueError(f"File {file} not supported.")

    @staticmethod
    def _read_srs(file):
        with file.open("rb") as buff:
            data = buff.read()
            if file.suffixes[-1] == ".gz":
                data = gzip.decompress(data)
        # Data is store as a series of records made of different numbers of bytes
        # General header information
        # 1		Year (last 2 digits)				Byte integer (unsigned)
        # 2		Month number (1 to 12)			    "
        # 3		Day (1 to 31)					    "
        # 4		Hour (0 to 23 UT)				    "
        # 5		Minute (0 to 59)				    "
        # 6		Second at start of scan (0 to 59)	"
        # 7		Site Number (0 to 255)			    "
        # 8		Number of bands in the record (2)	"
        #
        # Band 1 (A-band) header information
        # 9,10		Start Frequency (MHz)			    Word integer (16 bits)
        # 11,12		End Frequency (MHz)			        "
        # 13,14		Number of bytes in data record (401)"
        # 15		Analyser reference level		    Byte integer
        # 16		Analyser attenuation (dB)		    "
        #
        # Band 2 (B-band) header information
        # 17-24		As for band 1
        #
        # Spectrum Analyser data
        # 25-425	401 data bytes for band 1 (A-band)
        # 426-826	401 data bytes for band 2 (B-band)
        record_struc = struct.Struct("B" * 8 + "H" * 3 + "B" * 2 + "H" * 3 + "B" * 2 + "B" * 401 + "B" * 401)
        records = record_struc.iter_unpack(data)
        # Map of numeric records to locations
        site_map = {1: "Palehua", 2: "Holloman", 3: "Learmonth", 4: "San Vito"}
        df = pd.DataFrame([(*r[:18], np.array(r[18:419]), np.array(r[419:820])) for r in records])
        df.columns = [
            "year",
            "month",
            "day",
            "hour",
            "minute",
            "second",
            "site",
            " num_bands",
            "start_freq1",
            "end_freq1",
            "num_bytes1",
            "analyser_ref1",
            "analyser_atten1",
            "start_freq2",
            "end_freq2",
            "num_bytes2",
            "analyser_ref2",
            "analyser_atten2",
            "spec1",
            "spec2",
        ]
        # Hack to make to_datetime work - earliest dates seem to be 2000 and won't be
        # around in 3000!
        df["year"] = df["year"] + 2000
        df["time"] = pd.to_datetime(df[["year", "month", "day", "hour", "minute", "second"]])
        # Equations taken from document
        n = np.arange(1, 402)
        freq_a = (25 + 50 * (n - 1) / 400) * u.MHz
        freq_b = (75 + 105 * (n - 1) / 400) * u.MHz
        freqs = np.hstack([freq_a, freq_b])
        data = np.hstack([np.vstack(df[name].to_numpy()) for name in ["spec1", "spec2"]]).T
        times = Time(df["time"])
        meta = {
            "instrument": "RSTN",
            "observatory": site_map[df["site"][0]],
            "start_time": times[0],
            "end_time": times[-1],
            "detector": "RSTN",
            "wavelength": a.Wavelength(freqs[0], freqs[-1]),
            "freqs": freqs,
            "times": times,
        }
        return data, meta

    @staticmethod
    def _read_cdf(file):
        cdf = cdflib.CDF(file)

        cdf_globals = cdf.globalattsget()

        if (
            cdf_globals.get("Project", "")[0] == "PSP"
            and cdf_globals.get("Source_name")[0] == "PSP_FLD>Parker Solar Probe FIELDS"
            and "Radio Frequency Spectrometer" in cdf_globals.get("Descriptor")[0]
        ):
            short, _long = cdf_globals["Descriptor"][0].split(">")

            detector = short[4:].lower()
            times, data, freqs = [
                cdf.varget(name)
                for name in [
                    f"epoch_{detector}_auto_averages_ch0_V1V2",
                    f"psp_fld_l2_rfs_{detector}_auto_averages_ch0_V1V2",
                    f"frequency_{detector}_auto_averages_ch0_V1V2",
                ]
            ]
            times = Time("J2000.0", scale="tt") + (times << u.ns)
            freqs = freqs[0, :] << u.Hz
            data = data.T << u.Unit("Volt**2/Hz")
            meta = {
                "cdf_globals": cdf_globals,
                "detector": detector,
                "instrument": "FIELDS/RFS",
                "observatory": "PSP",
                "start_time": times[0],
                "end_time": times[-1],
                "wavelength": a.Wavelength(freqs.min(), freqs.max()),
                "times": times,
                "freqs": freqs,
            }
            return data, meta
        elif "SOLO" in cdf_globals.get("Project", "")[0]:
            if "RPW-HFR-SURV" not in cdf_globals.get("Descriptor", "")[0]:
                raise ValueError(
                    f"Currently radiospectra supports Level 2 HFR survey data the file"
                    f'{file.name} is {cdf_globals.get("Descriptor", "")}'
                )

            # FREQUENCY_BAND_LABELS = ["HF1", "HF2"]
            # SURVEY_MODE_LABELS = ["SURVEY_NORMAL", "SURVEY_BURST"]
            # CHANNEL_LABELS = ["1", "2"]
            SENSOR_MAPPING = {
                1: "V1",
                2: "V2",
                3: "V3",
                4: "V1-V2",
                5: "V2-V3",
                6: "V3-V1",
                7: "B_MF",
                9: "HF_V1-V2",
                10: "HF_V2-V3",
                11: "HF_V3-V1",
            }

            # Extract variables
            all_times = Time("J2000.0") + cdf.varget("EPOCH") * u.Unit(cdf.varattsget("EPOCH")["UNITS"])
            all_freqs = cdf.varget("FREQUENCY") << u.Unit(cdf.varattsget("FREQUENCY")["UNITS"])

            sweep_start_indices = np.asarray(np.diff(cdf.varget("SWEEP_NUM")) != 0).nonzero()
            sweep_start_indices = np.insert((sweep_start_indices[0] + 1), 0, 0)
            times = all_times[sweep_start_indices]

            sensor = cdf.varget("SENSOR_CONFIG")
            np.unique(cdf.varget("FREQUENCY"))
            band = cdf.varget("HFR_BAND")

            u.Unit(cdf.varattsget("AGC1").get("UNIT", "V^2/Hz"))
            agc1 = cdf.varget("AGC1")
            agc2 = cdf.varget("AGC2")

            # Define number of records
            n_rec = band.shape[0]
            # Get Epoch times of first sample of each sweep in the file
            sweep_times = times
            nt = len(sweep_times)
            # Get complete list of HFR frequency values
            hfr_frequency = 375 + 50 * np.arange(321)  # This is a guess something between 320 and 324
            nf = len(hfr_frequency)

            # Initialize output 2D array containing voltage spectral power values in V^2/Hz
            # Dims = (channels[2], time of the first sweep sample[len(time)], frequency[192])
            specs = np.empty((2, nt, nf))
            # Fill 2D array with NaN for HRF frequencies not actually measured in the file
            specs[:] = np.nan

            # Get list of first index of sweeps
            isweep = sweep_start_indices[:]
            # Get number of sweeps
            n_sweeps = len(isweep)
            # Insert an element in the end of the isweep list
            # containing the end of the latest sweep
            # (required for the loop below, in order to have
            # a start/end index range for each sweep)
            isweep = np.insert(isweep, n_sweeps, n_rec)

            # Initialize sensor_config
            sensor_config = np.zeros((2, nt), dtype=object)
            tm = []
            # Perform a loop on each sweep
            for i in range(n_sweeps):
                # Get first and last index of the sweep
                i0 = isweep[i]
                i1 = isweep[i + 1]

                ts = all_times[i0]
                te = all_times[i1 - 1]
                tt = (te - ts) * 0.5 + ts
                tm.append(tt)

                # Get indices of the actual frequency channels in the frequency vector
                freq_indices = ((all_freqs[i0:i1].value - 375) / 50).astype(int)

                # fill output 2D array
                specs[0, i, freq_indices] = agc1[i0:i1]
                specs[1, i, freq_indices] = agc2[i0:i1]

                # Fill sensor config
                sensor_config[0, i] = SENSOR_MAPPING[sensor[i0, 0]]
                sensor_config[1, i] = SENSOR_MAPPING[sensor[i0, 1]]

            # Define hfr bands
            hfc = np.array(["HF1", "HF2"])
            hfr_bands = hfc[band[:100] - 1]

            hfr_frequency = hfr_frequency << u.kHz

            res = []
            if np.any(agc1):
                meta1 = {
                    "cdf_globals": cdf_globals,
                    "detector": "RPW-AGC1",
                    "instrument": "RPW",
                    "observatory": "SOLO",
                    "start_time": times[0],
                    "end_time": times[-1],
                    "wavelength": a.Wavelength(hfr_frequency.min(), hfr_frequency.max()),
                    "times": times,
                    "freqs": hfr_frequency,
                }
                res.append((specs[0].T, meta1))
            if np.any(agc2):
                meta2 = {
                    "cdf_globals": cdf_globals,
                    "detector": "RPW-AGC2",
                    "instrument": "RPW",
                    "observatory": "SOLO",
                    "start_time": times[0],
                    "end_time": times[-1],
                    "wavelength": a.Wavelength(hfr_frequency.min(), hfr_frequency.max()),
                    "times": times,
                    "freqs": hfr_frequency,
                }
                res.append((specs[1].T, meta2))
            return res

    @staticmethod
    def _read_fits(file):
        hd_pairs = fits.open(file)
        if "e-CALLISTO" in hd_pairs[0].header.get("CONTENT", ""):
            data = hd_pairs[0].data
            times = hd_pairs[1].data["TIME"].flatten() * u.s
            freqs = hd_pairs[1].data["FREQUENCY"].flatten() * u.MHz
            start_time = parse_time(hd_pairs[0].header["DATE-OBS"] + " " + hd_pairs[0].header["TIME-OBS"])
            try:
                end_time = parse_time(hd_pairs[0].header["DATE-END"] + " " + hd_pairs[0].header["TIME-END"])
            except ValueError:
                # See https://github.com/sunpy/radiospectra/issues/74
                time_comps = hd_pairs[0].header["TIME-END"].split(":")
                time_comps[0] = "00"
                fixed_time = ":".join(time_comps)
                date_offset = parse_time(hd_pairs[0].header["DATE-END"] + " " + fixed_time)
                end_time = date_offset + 1 * u.day

            times = start_time + times
            meta = {
                "fits_meta": hd_pairs[0].header,
                "detector": "e-CALLISTO",
                "instrument": "e-CALLISTO",
                "observatory": hd_pairs[0].header["INSTRUME"],
                "start_time": start_time,
                "end_time": end_time,
                "wavelength": a.Wavelength(freqs.min(), freqs.max()),
                "times": times,
                "freqs": freqs,
            }
            return data, meta
        elif hd_pairs[0].header.get("TELESCOP", "") == "EOVSA":
            times = Time(hd_pairs[2].data["mjd"] + hd_pairs[2].data["time"] / 1000.0 / 86400.0, format="mjd")
            freqs = hd_pairs[1].data["sfreq"] * u.GHz
            data = hd_pairs[0].data
            start_time = parse_time(hd_pairs[0].header["DATE_OBS"])
            end_time = parse_time(hd_pairs[0].header["DATE_END"])
            meta = {
                "fits_meta": hd_pairs[0].header,
                "detector": "EOVSA",
                "instrument": "EOVSA",
                "observatory": "Owens Valley",
                "start_time": start_time,
                "end_time": end_time,
                "wavelength": a.Wavelength(freqs.min(), freqs.max()),
                "times": times,
                "freqs": freqs,
            }
            return data, meta
        # Semi standard - spec in primary and time and freq in 1st extension
        try:
            data = hd_pairs[0].data
            times = hd_pairs[1].data["TIME"].flatten() * u.s
            freqs = hd_pairs[1].data["FREQUENCY"].flatten() * u.MHz
            start_time = parse_time(hd_pairs[0].header["DATE-OBS"] + " " + hd_pairs[0].header["TIME-OBS"])
            end_time = parse_time(hd_pairs[0].header["DATE-END"] + " " + hd_pairs[0].header["TIME-END"])
            times = start_time + times
            meta = {
                "fits_meta": hd_pairs[0].header,
                "start_time": start_time,
                "end_time": end_time,
                "wavelength": a.Wavelength(freqs.min(), freqs.max()),
                "times": times,
                "freqs": freqs,
                "instrument": hd_pairs[0].header.get("INSTRUME", ""),
                "observatory": hd_pairs[0].header.get("INSTRUME", ""),
                "detector": hd_pairs[0].header.get("DETECTOR", ""),
            }
            if "e-CALLISTO" in hd_pairs[0].header["CONTENT"]:
                meta["detector"] = "e-CALLISTO"
                meta["instrument"] = "e-CALLISTO"
            return data, meta
        except Exception as e:
            raise ValueError(f"Could not load fits file: {file} into Spectrogram.") from e

    @staticmethod
    def _read_idl_sav(file, instrument=None):
        data = readsav(file)
        if instrument == "waves":
            # See https://solar-radio.gsfc.nasa.gov/wind/one_minute_doc.html
            data_array = data["arrayb"]
            # frequency range
            if file.suffix == ".R1":
                freqs = np.linspace(20, 1040, 256) * u.kHz
                receiver = "RAD1"
            elif file.suffix == ".R2":
                freqs = np.linspace(1.075, 13.825, 256) * u.MHz
                receiver = "RAD2"
            else:
                raise ValueError(f"Unknown WIND/WAVES file type: {file.suffix}")
            # bg which is already subtracted from data ?
            bg = data_array[:, -1]
            data = data_array[:, :-1]
            start_time = Time.strptime(file.stem, "%Y%m%d")
            end_time = start_time + 86399 * u.s
            times = start_time + (np.arange(1440) * 60 + 30) * u.s
            meta = {
                "instrument": "WAVES",
                "observatory": "WIND",
                "start_time": start_time,
                "end_time": end_time,
                "wavelength": a.Wavelength(freqs[0], freqs[-1]),
                "detector": receiver,
                "freqs": freqs,
                "times": times,
                "background": bg,
            }
            return data, meta
        else:
            raise ValueError(f"Unrecognized IDL .sav file: {file}")


Spectrogram = SpectrogramFactory(registry=GenericSpectrogram._registry, default_widget_type=GenericSpectrogram)



from radiospectra.net.attrs import *
from radiospectra.net.sources.ecallisto import eCALLISTOClient
from radiospectra.net.sources.eovsa import EOVSAClient
from radiospectra.net.sources.ilofar import ILOFARMode357Client
from radiospectra.net.sources.psp import RFSClient
from radiospectra.net.sources.rstn import RSTNClient

__all__ = [
    "eCALLISTOClient",
    "EOVSAClient",
    "RFSClient",
    "RSTNClient",
    "ILOFARMode357Client",
]



from radiospectra.spectrogram.spectrogrambase import GenericSpectrogram

__all__ = ["RFSSpectrogram"]


class RFSSpectrogram(GenericSpectrogram):
    """
    Parker Solar Probe FIELDS/Radio Frequency Spectrometer (RFS) Spectrogram.

    >>> import radiospectra.net
    >>> from sunpy.net import Fido, attrs as a
    >>> from radiospectra.spectrogram import Spectrogram
    >>> from radiospectra.net import attrs as ra
    >>> query = Fido.search(a.Time('2019/10/05 23:00', '2019/10/06 00:59'),  #doctest: +REMOTE_DATA
    ...                     a.Instrument.rfs)  #doctest: +REMOTE_DATA
    >>> downloaded = Fido.fetch(query[0][0])  #doctest: +REMOTE_DATA
    >>> spec = Spectrogram(downloaded[0])  #doctest: +REMOTE_DATA
    >>> spec  #doctest: +REMOTE_DATA
    <RFSSpectrogram PSP, FIELDS/RFS, LFR 10.546879882812501 kHz - 1687.5 kHz, 2019-10-05T00:01:32.395 to 2019-10-05T22:16:30.493>
    >>> spec.plot()  #doctest: +REMOTE_DATA
    <matplotlib.collections.QuadMesh object at ...>
    """

    def __init__(self, data, meta, **kwargs):
        super().__init__(meta=meta, data=data, **kwargs)

    @property
    def level(self):
        return self.meta["cdf_meta"]["Data_type"].split(">")[0]

    @property
    def version(self):
        return int(self.meta["cdf_meta"]["Data_version"])

    @classmethod
    def is_datasource_for(cls, data, meta, **kwargs):
        return (
            meta["observatory"] == "PSP" and meta["instrument"] == "FIELDS/RFS" and meta["detector"] in ("lfr", "hfr")
        )



import numpy as np

import astropy.units as u
from sunpy.net import attrs as a
from sunpy.net.dataretriever.client import GenericClient, QueryResponse
from sunpy.net.scraper import Scraper
from sunpy.time import TimeRange

from radiospectra.net.attrs import PolType

__all__ = ["ILOFARMode357Client"]

RECEIVER_FREQUENCIES = a.Wavelength(10.546875 * u.MHz, 244.53125 * u.MHz)
DATASET_NAMES = ["rcu357_1beam", "rcu357_1beam_datastream"]


class ILOFARMode357Client(GenericClient):
    """
    Provides access to I-LOFAR mode 357 observations from the
    data `archive <https://data.lofar.ie>`__

    Examples
    --------
    >>> import radiospectra.net
    >>> from sunpy.net import Fido, attrs as a
    >>> results = Fido.search(a.Time("2021/09/01", "2021/09/21"),
    ...                       a.Instrument('ILOFAR'))  # doctest: +REMOTE_DATA
    >>> results #doctest: +REMOTE_DATA
    <sunpy.net.fido_factory.UnifiedResponse object at ...>
    Results from 1 Provider:
    <BLANKLINE>
    10 Results from the ILOFARMode357Client:
    <BLANKLINE>
           Start Time               End Time        ... Provider Polarisation
    ----------------------- ----------------------- ... -------- ------------
    2021-09-14 07:39:13.000 2021-09-14 07:39:13.999 ...   ILOFAR            X
    2021-09-14 07:39:13.000 2021-09-14 07:39:13.999 ...   ILOFAR            Y
    2021-09-01 08:07:29.000 2021-09-01 08:07:29.999 ...   ILOFAR            X
    2021-09-01 08:07:29.000 2021-09-01 08:07:29.999 ...   ILOFAR            Y
    2021-09-07 08:07:52.000 2021-09-07 08:07:52.999 ...   ILOFAR            X
    2021-09-07 08:07:52.000 2021-09-07 08:07:52.999 ...   ILOFAR            Y
    2021-09-08 08:04:07.000 2021-09-08 08:04:07.999 ...   ILOFAR            X
    2021-09-08 08:04:07.000 2021-09-08 08:04:07.999 ...   ILOFAR            Y
    2021-09-08 10:34:31.000 2021-09-08 10:34:31.999 ...   ILOFAR            X
    2021-09-08 10:34:31.000 2021-09-08 10:34:31.999 ...   ILOFAR            Y
    <BLANKLINE>
    <BLANKLINE>
    """

    baseurl = r"https://data.lofar.ie/%Y/%m/%d/bst/kbt/{dataset}/" r"%Y%m%d_\d{{6}}_bst_00\S{{1}}.dat"

    pattern = r"{}/{year:4d}{month:2d}{day:2d}_{hour:2d}{minute:2d}{second:2d}" r"_bst_00{Polarisation}.dat"

    @classmethod
    def _check_wavelengths(cls, wavelength):
        """
        Check for overlap between given wavelength and receiver frequency coverage defined in
        `RECEIVER_FREQUENCIES`.

        Parameters
        ----------
        wavelength : `sunpy.net.attrs.Wavelength`
            Input wavelength range to check

        Returns
        -------
        `bool`
        """
        return wavelength.min in RECEIVER_FREQUENCIES or wavelength.max in RECEIVER_FREQUENCIES

    def search(self, *args, **kwargs):
        """
        Query this client for a list of results.

        Parameters
        ----------
        *args: `tuple`
            `sunpy.net.attrs` objects representing the query.
        **kwargs: `dict`
            Any extra keywords to refine the search.

        Returns
        -------
        A `QueryResponse` instance containing the query result.
        """
        matchdict = self._get_match_dict(*args, **kwargs)
        metalist = []

        wavelentgh = matchdict.get("Wavelength", False)
        if wavelentgh and not self._check_wavelengths(wavelentgh):
            return QueryResponse(metalist, client=self)

        tr = TimeRange(matchdict["Start Time"], matchdict["End Time"])

        for dataset in DATASET_NAMES:
            url = self.baseurl.format(dataset=dataset)
            scraper = Scraper(url, regex=True)
            filesmeta = scraper._extract_files_meta(tr, extractor=self.pattern)
            for i in filesmeta:
                rowdict = self.post_search_hook(i, matchdict)
                metalist.append(rowdict)

        query_response = QueryResponse(metalist, client=self)
        mask = np.full(len(query_response), True)
        pol = matchdict.get("PolType")
        if len(pol) == 1:
            pol = pol.upper()
            mask = mask & query_response["Polarisation"] == pol

        if query_response:
            query_response.remove_column("PolType")

        return query_response[mask]

    @classmethod
    def register_values(cls):
        adict = {
            a.Instrument: [("ILOFAR", "Irish LOFAR STATION (IE63)")],
            a.Source: [("ILOFAR", "Irish LOFAR Data Archive")],
            a.Provider: [("ILOFAR", "Irish LOFAR Data Archive")],
            a.Wavelength: [("*")],
            PolType: [("X", "X"), ("X Linear Polarisation", "Y Linear Polarisation")],
        }
        return adict



"""
Datasource-specific classes.

This is where datasource specific logic is implemented.
Each mission should have its own file with one or more classes defined.
"""

from ..spectrogram_factory import Spectrogram  # NOQA
from .callisto import *  # NOQA
from .eovsa import *  # NOQA
from .ilofar357 import *  # NOQA
from .psp_rfs import *  # NOQA
from .rpw import *  # NOQA
from .rstn import *  # NOQA
from .swaves import *  # NOQA
from .waves import *  # NOQA

__all__ = [
    "SWAVESSpectrogram",
    "RFSSpectrogram",
    "CALISTOSpectrogram",
    "EOVSASpectrogram",
    "RSTNSpectrogram",
    "RPWSpectrogram",
    "ILOFARMode357Spectrogram",
]



from pathlib import Path
from datetime import datetime
from unittest import mock

import numpy as np

import astropy.units as u
from astropy.time import Time
from sunpy.net import attrs as a

from radiospectra.spectrogram import Spectrogram
from radiospectra.spectrogram.sources import RSTNSpectrogram


@mock.patch("radiospectra.spectrogram.spectrogram_factory.parse_path")
def test_rstn(parse_path_moc):
    start_time = Time("2020-01-01T06:17:38.000")
    end_time = Time("2020-01-01T15:27:43.000")
    meta = {
        "instrument": "RSTN",
        "observatory": "San Vito",
        "start_time": start_time,
        "end_time": end_time,
        "detector": "RSTN",
        "wavelength": a.Wavelength(25000.0 * u.kHz, 180000.0 * u.kHz),
        "freqs": np.concatenate([np.linspace(25, 75, 401), np.linspace(75, 180, 401)]) * u.MHz,
        "times": start_time + np.linspace(0, (end_time - start_time).to_value("s"), 11003) * u.s,
    }
    array = np.zeros((802, 11003))
    parse_path_moc.return_value = [(array, meta)]
    file = Path("fakes.srs")
    spec = Spectrogram(file)
    assert isinstance(spec, RSTNSpectrogram)
    assert spec.observatory == "SAN VITO"
    assert spec.instrument == "RSTN"
    assert spec.detector == "RSTN"
    assert spec.start_time.datetime == datetime(2020, 1, 1, 6, 17, 38)
    assert spec.end_time.datetime == datetime(2020, 1, 1, 15, 27, 43)
    assert spec.wavelength.min == 25000 * u.kHz
    assert spec.wavelength.max == 180000 * u.kHz



import astropy.units as u

__all__ = ["subband_to_freq"]


def subband_to_freq(subband, obs_mode):
    """
    Converts LOFAR single station subbands to frequency

    Parameters
    ----------
    subband : `int`
        Subband number.
    obs_mode : `int`
        Observation mode 3, 5, 7.

    Return
    ------
    `astropy.units.Quantity`
        Frequency in MHz
    """
    nyq_zone_dict = {3: 1, 5: 2, 7: 3}
    if obs_mode not in nyq_zone_dict:
        raise ValueError(f"Observation mode {obs_mode} not supported, only 3, 5, 7 are supported.")
    nyq_zone = nyq_zone_dict[obs_mode]
    clock_dict = {3: 200, 4: 160, 5: 200, 6: 160, 7: 200}  # MHz
    clock = clock_dict[obs_mode]
    freq = (nyq_zone - 1 + subband / 512) * (clock / 2)
    return freq * u.MHz  # MHz



from pathlib import Path
from datetime import datetime
from unittest import mock

import numpy as np

import astropy.units as u
from astropy.time import Time
from sunpy.net import attrs as a

from radiospectra.spectrogram import Spectrogram
from radiospectra.spectrogram.sources import WAVESSpectrogram


@mock.patch("radiospectra.spectrogram.spectrogram_factory.parse_path")
def test_waves_rad1(parse_path_moc):
    meta = {
        "instrument": "WAVES",
        "observatory": "wind",
        "start_time": Time("2020-11-28 00:00:00"),
        "end_time": Time("2020-11-28 23:59:00"),
        "wavelength": a.Wavelength(20 * u.kHz, 1040 * u.kHz),
        "detector": "rad1",
        "freqs": np.linspace(20, 1040, 256) * u.kHz,
        "times": np.arange(1440) * u.min,
    }
    array = np.zeros((256, 1440))
    parse_path_moc.return_value = [(array, meta)]
    file = Path("fake.r1")
    spec = Spectrogram(file)
    assert isinstance(spec, WAVESSpectrogram)
    assert spec.observatory == "WIND"
    assert spec.instrument == "WAVES"
    assert spec.detector == "RAD1"
    assert spec.start_time.datetime == datetime(2020, 11, 28, 0, 0)
    assert spec.end_time.datetime == datetime(2020, 11, 28, 23, 59)
    assert spec.wavelength.min == 20.0 * u.kHz
    assert spec.wavelength.max == 1040.0 * u.kHz


@mock.patch("radiospectra.spectrogram.spectrogram_factory.parse_path")
def test_waves_rad2(parse_path_moc):
    meta = {
        "instrument": "WAVES",
        "observatory": "WIND",
        "start_time": Time("2020-11-28 00:00:00"),
        "end_time": Time("2020-11-28 23:59:00"),
        "wavelength": a.Wavelength(1.075 * u.MHz, 13.825 * u.MHz),
        "detector": "RAD2",
        "freqs": np.linspace(1.075, 13.825, 256) * u.MHz,
        "times": np.arange(1440) * u.min,
    }
    array = np.zeros((319, 1440))
    parse_path_moc.return_value = [(array, meta)]
    file = Path("fake.dat")
    spec = Spectrogram(file)
    assert isinstance(spec, WAVESSpectrogram)
    assert spec.observatory == "WIND"
    assert spec.instrument == "WAVES"
    assert spec.detector == "RAD2"
    assert spec.start_time.datetime == datetime(2020, 11, 28, 0, 0)
    assert spec.end_time.datetime == datetime(2020, 11, 28, 23, 59)
    assert spec.wavelength.min == 1.075 * u.MHz
    assert spec.wavelength.max == 13.825 * u.MHz



import gzip
from pathlib import Path
from unittest import mock

import numpy as np
import pytest

import astropy.units as u
from astropy.time import Time
from sunpy.net import Fido
from sunpy.net import attrs as a

from radiospectra.net.sources.psp import RFSClient

MOCK_PATH = "sunpy.net.scraper.urlopen"


@pytest.fixture
def client():
    return RFSClient()


@pytest.mark.parametrize(
    "req_wave,receivers",
    [
        # Completely contain the both receiver ranges
        (a.Wavelength(1 * u.kHz, 25000 * u.kHz), ["rfs_lfr", "rfs_hfr"]),
        # Min in lower freq and max in high freq receiver
        (a.Wavelength(20 * u.kHz, 15 * u.MHz), ["rfs_lfr", "rfs_hfr"]),
        # Min below and max in low freq receiver
        (a.Wavelength(1 * u.kHz, 100 * u.kHz), ["rfs_lfr"]),
        # Min and max in low freq receiver
        (a.Wavelength(20 * u.kHz, 100 * u.kHz), ["rfs_lfr"]),
        # Min and max in high freq receiver
        (a.Wavelength(1800 * u.kHz, 18000 * u.kHz), ["rfs_hfr"]),
        # Min in high freq receiver and max above
        (a.Wavelength(1800 * u.kHz, 20000 * u.kHz), ["rfs_hfr"]),
        # Min and max in the over lap
        (a.Wavelength(1.4 * u.MHz, 1.5 * u.MHz), ["rfs_lfr", "rfs_hfr"]),
    ],
)
def test_check_wavelength(req_wave, receivers, client):
    res = client._check_wavelengths(req_wave)
    assert set(res) == set(receivers)


@pytest.mark.remote_data
def test_fido():
    atr = a.Time("2019/10/01", "2019/10/02")
    res = Fido.search(atr, a.Instrument("rfs"))
    res0 = res[0]
    isinstance(res0.client, RFSClient)
    assert len(res0) == 4
    assert res["rfs"]["Start Time"].min() == Time("2019-10-01T00:00").datetime
    assert res["rfs"]["End Time"].max() == Time("2019-10-02T23:59:59.999").datetime


@pytest.fixture
def http_responces():
    paths = [Path(__file__).parent / "data" / n for n in ["psp_resp1.html.gz", "psp_resp2.html.gz"]]
    response_htmls = []
    for p in paths:
        with gzip.open(p) as f:
            response_htmls.append(f.read())
    return response_htmls


@mock.patch(MOCK_PATH)
def test_search_with_wavelength(mock_urlopen, client, http_responces):
    mock_urlopen.return_value.read = mock.MagicMock()
    mock_urlopen.return_value.read.side_effect = http_responces
    mock_urlopen.close = mock.MagicMock(return_value=None)
    tr = a.Time("2019/10/13", "2019/10/15")
    wr1 = a.Wavelength(1 * u.kHz, 1.1 * u.MHz)
    res1 = client.search(tr, wr1)

    mock_urlopen.assert_called_with("https://spdf.gsfc.nasa.gov/pub/data/psp/fields/l2/rfs_lfr/2019/")
    assert np.array_equal(res1[0]["Wavelength"], [10, 1700] * u.kHz)
    assert len(res1) == 3
    assert res1["Start Time"].min().datetime == Time("2019-10-13T00:00").datetime
    assert res1["End Time"].max().datetime == Time("2019-10-15T23:59:59.999").datetime
    wr2 = a.Wavelength(2 * u.MHz, 20 * u.MHz)
    res2 = client.search(tr, wr2)
    mock_urlopen.assert_called_with("https://spdf.gsfc.nasa.gov/pub/data/psp/fields/l2/rfs_hfr/2019/")
    assert np.array_equal(res2[0]["Wavelength"], [1300, 19200] * u.kHz)
    assert len(res2) == 3
    assert res2.time_range().start == Time("2019-10-13T00:00").datetime
    assert res2.time_range().end == Time("2019-10-15T23:59:59.999").datetime


@pytest.mark.remote_data
def test_get_url_for_time_range(client):
    url_start = "https://spdf.gsfc.nasa.gov/pub/data/psp/fields/l2/rfs_lfr/2019/" "psp_fld_l2_rfs_lfr_20191001_v03.cdf"
    url_end = "https://spdf.gsfc.nasa.gov/pub/data/psp/fields/l2/rfs_hfr/2019/" "psp_fld_l2_rfs_hfr_20191015_v03.cdf"
    tr = a.Time("2019/10/01", "2019/10/15")
    res = client.search(tr)
    urls = [i["url"] for i in res]
    assert urls[0] == url_start
    assert urls[-1] == url_end


def test_can_handle_query(client):
    atr = a.Time("2019/10/01", "2019/11/01")
    res = client._can_handle_query(atr, a.Instrument("rfs"))
    assert res is True
    res = client._can_handle_query(atr)
    assert res is False


@pytest.mark.remote_data
def test_download(client):
    query = client.search(a.Time("2019/10/05", "2019/10/06"), a.Instrument("rfs"))
    download_list = client.fetch(query)
    assert len(download_list) == len(query)



from pathlib import Path
from unittest import mock

import pytest

from sunpy.net import attrs as a
from sunpy.net.fido_factory import Fido

from radiospectra.net.attrs import Observatory
from radiospectra.net.sources.rstn import RSTNClient

MOCK_PATH = "sunpy.net.scraper.urlopen"


@pytest.fixture
def client():
    return RSTNClient()


@pytest.fixture
def http_responses():
    paths = [
        Path(__file__).parent / "data" / n for n in ["rstn_holloman.html", "rstn_learmonth.html", "rstn_san-vito.html"]
    ]
    response_htmls = []
    for p in paths:
        with p.open("r") as f:
            response_htmls.append(f.read())

    # For the chosen test dates there was no data form  Palehua or Sagamore so insert two empty
    # responses in there place
    response_htmls = response_htmls[:2] + ["", ""] + [response_htmls[-1]]
    return response_htmls


@mock.patch(MOCK_PATH)
def test_client(urlopen, client, http_responses):
    urlopen.return_value.read = mock.MagicMock()
    urlopen.return_value.read.side_effect = http_responses
    urlopen.close = mock.MagicMock(return_value=None)
    query = client.search(a.Time("2003/03/15 00:00", "2003/03/15 23:59"), a.Instrument("RSTN"))
    assert urlopen.call_count == 5
    # last call arg should be san-vito url
    assert (
        urlopen.call_args[0][0] == "https://www.ngdc.noaa.gov/stp/space-weather/solar-data/"
        "solar-features/solar-radio/rstn-spectral/san-vito/2003/03/"
    )
    assert len(query) == 3


@mock.patch(MOCK_PATH)
def test_client_observatory(urlopen, client, http_responses):
    urlopen.return_value.read = mock.MagicMock()
    urlopen.return_value.read.side_effect = http_responses[-1:]
    urlopen.close = mock.MagicMock(return_value=None)
    query = client.search(a.Time("2003/03/15 00:00", "2003/03/15 23:59"), a.Instrument("RSTN"), Observatory("San Vito"))
    urlopen.assert_called_once_with(
        "https://www.ngdc.noaa.gov/stp/space-weather/solar-data/"
        "solar-features/solar-radio/rstn-spectral/san-vito/2003/03/"
    )
    assert len(query) == 1
    assert query["Observatory"] == "San Vito"


@pytest.mark.remote_data
def test_fido():
    query = Fido.search(a.Time("2003/03/15 00:00", "2003/03/15 23:59"), a.Instrument("RSTN"), Observatory("San Vito"))
    assert len(query[0]) == 1
    assert all(query[0]["Observatory"] == "San Vito")



from radiospectra.spectrogram.spectrogrambase import GenericSpectrogram

__all__ = ["RPWSpectrogram"]


class RPWSpectrogram(GenericSpectrogram):
    """
    Solar Orbiter Radio and Plasma Waves (RPW) RPW-HFR-SURV spectrogram.

    For more information on the instrument see `<https://rpw.lesia.obspm.fr>`__.

    Examples
    --------
    >>> import sunpy_soar
    >>> from sunpy.net import Fido, attrs as a
    >>> from radiospectra.spectrogram import Spectrogram
    >>> query = Fido.search(a.Time('2020-07-11', '2020-07-11 23:59'), a.Instrument('RPW'),
    ...             a.Level(2), a.soar.Product('RPW-HFR-SURV')) #doctest: +REMOTE_DATA
    >>> downloaded = Fido.fetch(query[0])  #doctest: +SKIP
    >>> spec = Spectrogram(downloaded[0])  #doctest: +SKIP
    >>> spec  #doctest: +SKIP
    [<RPWSpectrogram SOLO, RPW, RPW-AGC1 375.0 kHz - 16375.0 kHz, 2020-07-11T00:00:39.352 to 2020-07-12T00:00:55.715>, <RPWSpectrogram SOLO, RPW, RPW-AGC2 375.0 kHz - 16375.0 kHz, 2020-07-11T00:00:39.352 to 2020-07-12T00:00:55.715>]
    >>> spec[0] .plot()  #doctest: +SKIP
    <matplotlib.collections.QuadMesh object at ...>
    """

    @classmethod
    def is_datasource_for(cls, data, meta, **kwargs):
        return meta["instrument"] == "RPW"



from radiospectra.spectrogram.spectrogrambase import GenericSpectrogram

__all__ = ["WAVESSpectrogram"]


class WAVESSpectrogram(GenericSpectrogram):
    """
    Wind Waves Spectrogram.

    Examples
    --------
    >>> import radiospectra.net  #doctest: +SKIP
    >>> from sunpy.net import Fido, attrs as a  #doctest: +SKIP
    >>> from radiospectra.spectrogram import Spectrogram  #doctest: +SKIP
    >>> from radiospectra.net import attrs as ra  #doctest: +SKIP
    >>> query = Fido.search(a.Time('2019/10/05 23:00', '2019/10/06 00:59'),  #doctest: +REMOTE_DATA +SKIP
    ...                     a.Instrument('WAVES'))  #doctest: +REMOTE_DATA +SKIP
    >>> downloaded = Fido.fetch(query[0][0])  #doctest: +REMOTE_DATA +SKIP
    >>> spec = Spectrogram(downloaded[0])  #doctest: +REMOTE_DATA +SKIP
    >>> spec  #doctest: +REMOTE_DATA +SKIP
    <WAVESSpectrogram WIND, WAVES, RAD1 20.0 kHz - 1040.0 kHz, 2019-10-05T00:00:00.000 to 2019-10-05T23:59:59.000>
    >>> spec.plot()  #doctest: +REMOTE_DATA +SKIP
    <matplotlib.collections.QuadMesh object at ...>
    """

    def __init__(self, data, meta, **kwargs):
        super().__init__(meta=meta, data=data, **kwargs)

    @property
    def receiver(self):
        """
        The name of the receiver.
        """
        return self.meta["receiver"]

    @property
    def background(self):
        """
        The background subtracted from the data.
        """
        return self.meta.bg

    @classmethod
    def is_datasource_for(cls, data, meta, **kwargs):
        return meta.get("instrument", None) == "WAVES"



import gzip
from pathlib import Path
from unittest import mock

import pytest

from sunpy.net import attrs as a
from sunpy.net.fido_factory import Fido

from radiospectra.net.sources.ecallisto import Observatory, eCALLISTOClient

MOCK_PATH = "sunpy.net.scraper.urlopen"


@pytest.fixture
def client():
    return eCALLISTOClient()


@pytest.fixture
def http_responses():
    paths = [Path(__file__).parent / "data" / n for n in ["ecallisto_resp1.html.gz", "ecallisto_resp2.html.gz"]]
    response_htmls = []
    for p in paths:
        with gzip.open(p) as f:
            response_htmls.append(f.read())
    return response_htmls


@pytest.fixture
def http_response_alt():
    path = Path(__file__).parent / "data" / "ecallisto_resp_alt_format.html"
    with path.open("r") as file:
        response_html = file.read()
        return response_html


@mock.patch(MOCK_PATH)
def test_client(urlopen, client, http_responses):
    urlopen.return_value.read = mock.MagicMock()
    urlopen.return_value.read.side_effect = http_responses
    urlopen.close = mock.MagicMock(return_value=None)
    query = client.search(a.Time("2019/10/05 23:00", "2019/10/06 00:59"), a.Instrument("eCALLISTO"))
    assert urlopen.call_count == 2
    # 2nd call
    urlopen.assert_called_with("http://soleil80.cs.technik.fhnw.ch/solarradio/data/2002-20yy_Callisto/2019/10/06/")
    assert len(query) == 156


@mock.patch(MOCK_PATH)
def test_client_alt_format(urlopen, client, http_response_alt):
    urlopen.return_value.read = mock.MagicMock()
    urlopen.return_value.read.return_value = http_response_alt
    urlopen.close = mock.MagicMock(return_value=None)
    query = client.search(
        a.Time("2010/03/27 00:00", "2010/03/27 23:59"), a.Instrument("eCALLISTO"), Observatory("PHOENIX3-B1")
    )
    assert len(query) == 24


@mock.patch(MOCK_PATH)
def test_client_with_observatory(urlopen, client, http_responses):
    urlopen.return_value.read = mock.MagicMock()
    urlopen.return_value.read.side_effect = http_responses
    urlopen.close = mock.MagicMock(return_value=None)
    query = client.search(
        a.Time("2019/10/05 23:00", "2019/10/06 00:59"), a.Instrument("eCALLISTO"), Observatory("ALASKA")
    )
    assert urlopen.call_count == 2
    # 2nd call
    urlopen.assert_called_with("http://soleil80.cs.technik.fhnw.ch/solarradio/data/2002-20yy_Callisto/2019/10/06/")
    assert len(query) == 8


@pytest.mark.remote_data
def test_fido():
    query = Fido.search(
        a.Time("2019/10/05 23:00", "2019/10/06 00:59"), a.Instrument("eCALLISTO"), Observatory("ALASKA")
    )
    assert len(query[0]) == 8
    assert all(query[0]["Observatory"] == "ALASKA")



from pathlib import Path
from unittest import mock

import numpy as np

from radiospectra.spectrogram import Spectrogram


@mock.patch("numpy.fromfile")
@mock.patch("sunpy.util.io.is_file")
def test_ilofar(mock_is_file, mock_fromfile):
    mock_fromfile.return_value = np.ones(10117216)
    mock_is_file.return_value = True

    spec = Spectrogram(Path("20180602_063247_bst_00X.dat"))
    assert len(spec) == 3
    assert spec[0].polarisation == "X"
    assert spec[0].start_time.iso == "2018-06-02 06:32:47.000"
    assert spec[0].end_time.iso == "2018-06-02 12:18:18.000"
    assert spec[0].mode == 3
    assert spec[0].frequencies[0].to_value("MHz") == 10.546875
    assert spec[0].frequencies[-1].to_value("MHz") == 88.28125
    assert spec[1].mode == 5
    assert spec[1].frequencies[0].to_value("MHz") == 110.546875
    assert spec[1].frequencies[-1].to_value("MHz") == 188.28125
    assert spec[2].mode == 7
    assert spec[2].frequencies[0].to_value("MHz") == 210.546875
    assert spec[2].frequencies[-1].to_value("MHz") == 244.53125



from radiospectra.exceptions import SpectraMetaValidationError
from radiospectra.mixins import NonUniformImagePlotMixin, PcolormeshPlotMixin

__all__ = ["GenericSpectrogram"]


class GenericSpectrogram(PcolormeshPlotMixin, NonUniformImagePlotMixin):
    """
    Base spectrogram class all spectrograms inherit.

    Attributes
    ----------
    meta : `dict-like`
        Meta data for the spectrogram.
    data : `numpy.ndarray`
        The spectrogram data itself a 2D array.
    """

    _registry = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "is_datasource_for"):
            cls._registry[cls] = cls.is_datasource_for

    def __init__(self, data, meta, **kwargs):
        self.data = data
        self.meta = meta
        self._validate_meta()

    @property
    def observatory(self):
        """
        The name of the observatory which recorded the spectrogram.
        """
        return self.meta["observatory"].upper()

    @property
    def instrument(self):
        """
        The name of the instrument which recorded the spectrogram.
        """
        return self.meta["instrument"].upper()

    @property
    def detector(self):
        """
        The detector which recorded the spectrogram.
        """
        return self.meta["detector"].upper()

    @property
    def start_time(self):
        """
        The start time of the spectrogram.
        """
        return self.meta["start_time"]

    @property
    def end_time(self):
        """
        The end time of the spectrogram.
        """
        return self.meta["end_time"]

    @property
    def wavelength(self):
        """
        The wavelength range of the spectrogram.
        """
        return self.meta["wavelength"]

    @property
    def times(self):
        """
        The times of the spectrogram.
        """
        return self.meta["times"]

    @property
    def frequencies(self):
        """
        The frequencies of the spectrogram.
        """
        return self.meta["freqs"]

    def _validate_meta(self):
        """
        Validates the meta-information associated with a Spectrogram.

        This method includes very basic validation checks which apply to
        all of the kinds of files that radiospectra can read.
        Datasource-specific validation should be handled in the relevant
        file in the radiospectra.spectrogram.sources.
        """
        msg = "Spectrogram coordinate units for {} axis not present in metadata."
        err_message = []
        for i, ax in enumerate(["times", "freqs"]):
            if self.meta.get(ax) is None:
                err_message.append(msg.format(ax))
        if err_message:
            raise SpectraMetaValidationError("\n".join(err_message))

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} {self.observatory}, {self.instrument}, {self.detector}"
            f" {self.wavelength.min} - {self.wavelength.max},"
            f" {self.start_time.isot} to {self.end_time.isot}>"
        )



from sunpy.net import attrs as a
from sunpy.net.attr import SimpleAttr
from sunpy.net.dataretriever.client import GenericClient

from radiospectra.net.attrs import Observatory


class eCALLISTOClient(GenericClient):
    """
    Provides access to `eCallisto radio spectrometer <http://soleil80.cs.technik.fhnw.ch/solarradio/data/2002-20yy_Callisto/>`__
    `data archive <https://spdf.gsfc.nasa.gov>`__.

    `Further information <http://www.e-callisto.org>`__.

    Notes
    -----
    `Specific information on the meaning of the filename. <http://soleil.i4ds.ch/solarradio/data/readme.txt>`__

    From the filename alone there's no way to tell about either the frequency or duration.
    Therefore we only return a start time.

    Examples
    --------
    >>> from radiospectra import net
    >>> from sunpy.net import Fido, attrs as a
    >>> query = Fido.search(a.Time('2019/10/05 23:00', '2019/10/05 23:30'),
    ...                     a.Instrument('eCALLISTO'), net.Observatory('ALASKA'))  #doctest: +REMOTE_DATA
    >>> query  #doctest: +REMOTE_DATA
    <sunpy.net.fido_factory.UnifiedResponse object at ...>
    Results from 1 Provider:
    <BLANKLINE>
    3 Results from the eCALLISTOClient:
           Start Time        Provider Instrument Observatory  ID
    ----------------------- --------- ---------- ----------- ---
    2019-10-05 23:00:00.000 ECALLISTO  ECALLISTO      ALASKA  59
    2019-10-05 23:15:00.000 ECALLISTO  ECALLISTO      ALASKA  59
    2019-10-05 23:30:00.000 ECALLISTO  ECALLISTO      ALASKA  59
    <BLANKLINE>
    <BLANKLINE>
    """

    baseurl = (
        r"http://soleil80.cs.technik.fhnw.ch/solarradio/data/2002-20yy_Callisto/"
        r"%Y/%m/%d/{obs}_%Y%m%d_%H%M%S.*.fit.gz"
    )
    pattern = (
        r"{}/2002-20yy_Callisto/{year:4d}/{month:2d}/{day:2d}/"
        r"{Observatory}_{year:4d}{month:2d}{day:2d}"
        r"_{hour:2d}{minute:2d}{second:2d}{suffix}.fit.gz"
    )

    @classmethod
    def pre_search_hook(cls, *args, **kwargs):
        baseurl, pattern, matchdict = super().pre_search_hook(*args, **kwargs)
        obs = matchdict.pop("Observatory")
        if obs[0] == "*":
            baseurl = baseurl.format(obs=r".*")
        else:
            # Need case sensitive so have to override
            obs_attr = [a for a in args if isinstance(a, Observatory)][0]
            baseurl = baseurl.format(obs=obs_attr.value)
        return baseurl, pattern, matchdict

    def post_search_hook(self, exdict, matchdict):
        original = super().post_search_hook(exdict, matchdict)
        original["ID"] = original["suffix"].replace("_", "")
        del original["suffix"]
        # We don't know the end time for all files
        # https://github.com/sunpy/radiospectra/issues/60
        del original["End Time"]
        return original

    @classmethod
    def register_values(cls):
        adict = {
            a.Provider: [("eCALLISTO", "International Network of Solar Radio Spectrometers.")],
            a.Instrument: [("eCALLISTO", "e-Callisto - International Network of Solar Radio Spectrometers.")],
            Observatory: [("*", "Observatory Location")],
        }
        return adict

    @classmethod
    def _can_handle_query(cls, *query):
        """
        Method the `sunpy.net.fido_factory.UnifiedDownloaderFactory` class uses
        to dispatch queries to this Client.
        """
        regattrs_dict = cls.register_values()
        optional = {k for k in regattrs_dict.keys()} - cls.required
        if not cls.check_attr_types_in_query(query, cls.required, optional):
            return False
        for key in regattrs_dict:
            all_vals = [i[0].lower() for i in regattrs_dict[key]]
            for x in query:
                if (
                    isinstance(x, key)
                    and issubclass(key, SimpleAttr)
                    and x.type_name != "observatory"
                    and str(x.value).lower() not in all_vals
                ):
                    return False
        return True



"""
================
``radiospectra``
================

An open-source Python library for radio spectra in solar physics.

* Homepage: https://sunpy.org
* Documentation: https://docs.radiospectra.org/en/stable/
* Source code: https://github.com/sunpy/radiospectra
"""

from .version import version as __version__

__all__ = ["__version__"]



from radiospectra.spectrogram.spectrogrambase import GenericSpectrogram

__all__ = ["EOVSASpectrogram"]


class EOVSASpectrogram(GenericSpectrogram):
    """
    Extend Owen Valley Array (EOVSA) Spectrogram.

    Examples
    --------
    >>> import radiospectra.net
    >>> from sunpy.net import Fido, attrs as a
    >>> from radiospectra.spectrogram import Spectrogram
    >>> query = Fido.search(a.Time('2021/05/07 00:00', '2021/05/07 23:00'), a.Instrument.eovsa)  #doctest: +REMOTE_DATA
    >>> downloaded = Fido.fetch(query[0][0])  #doctest: +REMOTE_DATA
    >>> spec = Spectrogram(downloaded[0])  #doctest: +REMOTE_DATA
    >>> spec  #doctest: +REMOTE_DATA
    <EOVSASpectrogram OWENS VALLEY, EOVSA, EOVSA 1105371.117591858 kHz - 17979686.737060547 kHz, 2021-05-07T13:48:20.999 to 2021-05-08T01:50:59.999>
    >>> spec.plot()  #doctest: +REMOTE_DATA
    <matplotlib.collections.QuadMesh object at ...>
    """

    def __init__(self, data, meta, **kwargs):
        super().__init__(meta=meta, data=data, **kwargs)

    @property
    def polarisation(self):
        return self.meta["fits_meta"]["POLARIZA"]

    @classmethod
    def is_datasource_for(cls, data, meta, **kwargs):
        return meta["instrument"] == "EOVSA" or meta["detector"] == "EOVSA"

    # TODO fix time gaps for plots need to render them as gaps
    # can prob do when generateing proper pcolormesh gird but then prob doesn't belong here



import numpy as np

from radiospectra.spectrum import Spectrum


def test_spectrum():
    spec = Spectrum(np.arange(10), np.arange(10))
    np.testing.assert_equal(spec.data, np.arange(10))
    np.testing.assert_equal(spec.freq_axis, np.arange(10))
